@@ -26,7 +26,6 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#include <sys/cdefs.h>
 
 /*
  * xdr_stdio.c, XDR implementation on standard i/o file.
@@ -38,7 +37,6 @@
  * from the stream.
  */
 
-#include "namespace.h"
 #include <stdio.h>
 
 #include <arpa/inet.h>
